__name__ = "nanogram.filters"
from .types import types
import re


class Filter:
    """
        A filter for the handlers.

        Note: Use this class as a base
        class, because the verify method
        does nothing by default
    """
    def __init__(self):
        pass

    def verify(self, _type):
        """
            Checks if the Filter conditions
            are true
        """
        return True, self


class Regex(Filter):
    """
        A regex Filter for the handlers.

        Based on <Filter> class
    """
    def __init__(self, regex, use_search=True):
        self.regex = regex
        self.use_search = use_search

    def verify(self, _type):
        if "text" in _type.keys():
            if self.use_search:
                return isinstance(
                    re.search(self.regex, _type["text"]), re.Match
                )

            else:
                return isinstance(
                    re.match(self.regex, _type["text"]), re.Match
                )
        return False
