__name__ = "nanogram.handlers"
from typing import Callable, Union
from .types.types import types
from .filters import Filter


class Handler:
    def __init__(
        self,
        _filter: Union[Filter, str],
        callback: Callable,
        support_groups: bool = True,
        priority: int = -1,
        accept_types: list[str] = ["Message"]
    ):
        if _filter is None:
            raise ValueError(
                "_filter argument can't be None"
            )

        if callback is None:
            raise ValueError(
                "callback argument can't be None"
            )

        self._filter = _filter
        self.callback = callback
        self.support_groups = support_groups

        for i in accept_types:
            if i not in types:
                raise IndexError(f"Invalid type {i}")

        self.accept_types = accept_types
        self.priority = priority

    def run(self, bot, update):
        if self.support_groups is False:
            if update["chat"]["type"] in ["supergroup", "group"]:
                return

        if type(self._filter) is not str:
            if self._filter.verify(update):
                self.callback(bot, update)

        else:
            if hasattr(update, "text"):
                if self._filter == update["text"]:
                    self.callback(bot, update)

            else:
                return


class CommandHandler(Handler):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def run(self, bot, update):
        if "entities" not in update.keys():
            return

        if self.support_groups is False:
            if update["chat"]["type"] in ["supergroup", "group"]:
                return

        command = None
        for i in update["entities"]:
            if i["type"] == "bot_command":
                command = update["text"][0: i["length"]]
                break

        if command is None:
            return

        if type(self._filter) == str:
            if command == self._filter \
                    if self._filter[0] == "/" \
                    else f"/{self._filter}":
                self.callback(bot, update)

        else:
            if self._filter.verify(command):
                self.callback(bot, update)


class MessageHandler(Handler):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def run(self, bot, update):
        if update.type == "Message":
            super().run(bot, update)

        else:
            return
