__name__ = "nanogram.polling"
from .types.dynamic import DynamicType
from .handlers import Handler


class PollingUpdater:
    """
        The built-in polling updater
    """
    def __init__(self, bot_instance):
        self.bot = bot_instance
        self.handlers = {
            "all": {
                "-1": []
            }
        }

    def clear_updates(self):
        """
            Clear's the update's which
            are in the telegram queue
        """
        updates = self.bot.getUpdates(timeout=1)
        if len(updates) == 0:
            return updates

        updates = self.bot.getUpdates(
            offset=updates[
                len(updates) - 1
            ]["update_id"] + 1
        )

        return updates

    def add_handler(self, handler: Handler):
        """
            Add's an handler to the handler's list
        """
        priority = handler.priority if hasattr(handler, "priority") else -1
        _types = handler.accept_types \
            if hasattr(handler, "accept_types") \
            else ["all"]

        for i in _types:
            if i not in [*self.handlers.keys()]:
                self.handlers[i] = {}

            if priority not in [*self.handlers[i].keys()]:
                self.handlers[i][priority] = []

            self.handlers[i][priority].append(handler)

    def add_handlers(self, handlers: list[Handler]):
        """
            Register's a list of Handler's
        """
        for handler in handlers:
            self.add_handler(handler)

    def start_polling(self, **kwargs):
        """
            Start's polling
        """
        if "timeout" not in kwargs:
            kwargs["timeout"] = 60

        update_id = 0
        while True:
            try:
                updates = self.bot.getUpdates(**kwargs, offset=update_id + 1)
                for update in updates:
                    data = DynamicType(update, "Update")
                    update_id = data["update_id"]
                    del data["update_id"]

                    for i in data:
                        _type = DynamicType(i)
                        if _type.type not in [*self.handlers.keys()]:
                            pass

                        else:
                            prioritys = [
                                *self.handlers[_type.type].keys()
                            ]
                            prioritys.sort()
                            prioritys.reverse()
                            for prior in prioritys:
                                for handler in self.handlers[
                                    _type.type
                                ][prior]:
                                    handler.run(self.bot, _type)

            except Exception as e:
                raise e
