__name__ = "nanogram.types.dynamic"
from .types import types, telegram_type, optimized_types
from typing import Union


class DynamicType:
    """
        A telegram type

        Example:
            >>> import nanogram.types.dynamic
            >>> type = nanogram.types.dynamic.DynamicType({
            ...     "update_id": 0
            ...})
    """
    def __init__(self, data, _type=None):
        self.type = None
        __data__ = data.copy()

        if _type is None:
            # Iter's the type's and verify the len of them
            for _type in optimized_types:
                if _type < len(__data__) + 1:
                    for _subtype in optimized_types[_type]:
                        if (
                            optimized_types[_type][_subtype][0] + _type
                        ) < len(__data__):
                            continue

                        try:
                            verifyType(__data__, _subtype)
                            self.type = _subtype
                            break

                        except Exception as e:
                            pass

                    if self.type is None:
                        continue

                    break

        else:
            verifyType(__data__, _type)
            self.type = _type

        if self.type is None:
            raise Exception("Type not founded")

        self.__name__ = self.type
        self.data = __data__

    def values(self):
        """Return's the structure values"""
        return [*self.data.values()]

    def keys(self):
        """Return's the structure keys"""
        return [*self.data.keys()]

    def __repr__(self):
        return f"<{self.type} {self.data}>"

    def __getattr__(self, arg):
        return self.data[arg]

    def __getitem__(self, arg):
        if isinstance(arg, int):
            return [*self.values()][arg]
        return self.data[arg]

    def __delitem__(self, arg):
        del self.data[arg]


def checkn(t, c, b):
    for i in c:
        if i not in b:
            raise IndexError(f"Invalid Key '{i}'")

        if (
            b[i] if isinstance(b[i], str) else ""
        ) in telegram_type:
            verifyType(c[i], b[i])

        elif (
            (isinstance(c[i], b[i])) if isinstance(b[i], Union[str, list]) is False else None
        ):
            pass

        elif isinstance(b[i], list):
            if b[i][0] == telegram_type["array"]:
                if b[i][1] == telegram_type["array"]:
                    for child in c[i]:
                        for subchild in child:
                            verifyType(subchild, b[i][1][1])

                else:
                    for child in c[i]:
                        verifyType(child, b[i][1])

        else:
            raise ValueError(
                f"Invalid data for '{i}', {b[i]} was needed, not {type(c[i])}."
            )


def checkw(t, c, b):
    for i in b:
        if i not in c:
            raise IndexError(f"Key '{i}' its required for '{t}'")

        if (
            b[i] if isinstance(b[i], str) else ""
        ) in telegram_type:
            verifyType(c[i], b[i])

        elif (
            isinstance(c[i], b[i]) if isinstance(b[i], Union[str, list]) is False else None
        ):
            pass

        elif isinstance(b[i], list):
            if b[i][0] == telegram_type["array"]:
                print(b[i][1])
                if b[i][1] == telegram_type["array"]:
                    for child in c[i]:
                        for subchild in child:
                            verifyType(subchild, b[i][1][1])

                elif isinstance(b[i][1], list):
                    for child in c[i][1]:
                        verifyType(child, b[i][1][1])

                else:
                    for child in c[i]:
                        verifyType(child, b[i][1])

        else:
            raise ValueError(
                f"Invalid data for '{i}', {b[i]} was needed, not {type(c[i])}."
            )
        c.pop(i)

    return c


def verifyType(data, _type):
    if _type not in types:
        raise IndexError(f"Type {_type} not founded!")

    data_ = types[_type]

    data = checkw(_type, data.copy(), data_[0])

    if len(data_) != 1:
        checkn(_type, data.copy(), data_[1])

    return True
